
/* 
 * Codifica di Huffman (6/06/2021)
 *
 * Classe Huffman: algoritmi di compressione e decompressione
 */

// import java.util.*;

import campus_toolkit.*;


public class HuffmanCompression {


  private static final int BYTES = InputByteFile.BYTES;
  
  
  // ----------  Compressione  ----------
  
  
  // 1.
  // Istogramma delle frequenze dei byte
  // src: nome del documento da comprimere
  
  public static int[] byteHistogram( String src ) {
  
    InputByteFile in = new InputByteFile( src );
    
    int[] freq = new int[ BYTES ];                          // istogramma delle frequenze dei caratteri
    
    for ( int b=0; b<BYTES; b=b+1 ) {                       // inizializzazione istogramma
      freq[b] = 0;
    }
    
    while ( in.bytesAvailable() ) {                         // scansione: frequenze dei caratteri
    
      int b = in.readByte();                                // (ulteriore) occorrenza di b
      freq[b] = freq[b] + 1;
    }
    in.close();
    
    return freq;
  }
  
  
  // 2.
  // Albero binario di ri-codifica dei byte
  // freq: istogramma delle frequenze
  
  public static Node huffmanTree( int[] freq ) {
  
    NodeQueue queue = new NodeQueue();                      // coda con priorità di nodi
    
    for ( int b=0; b<BYTES; b=b+1 ) {                       // per ogni byte utilizzato
      if ( freq[b] > 0 ) {
        Node n = new Node( b, freq[b] );                    // creazione del nodo corrispondente
        queue.add( n );                                     // e inserimento nella coda
    }}
    while ( queue.size() > 1 ) {                            // finché c'è più di un nodo nella coda
    
      Node l = queue.poll();                                // rimozione dei due nodi di peso minimo...
      Node r = queue.poll();
      
      Node n = new Node( l, r );                            // ...costruzione di un (sotto) albero
      queue.add( n );                                       // e inserimento nella coda
    }
    return queue.poll();                                    // radice dell'albero di Huffman
  }
  
  
  // 3.
  // Tabella di ri-codifica dei byte:
  // root: nodo radice dell'albero di Huffman
  
  public static String[] huffmanCodesTable( Node root ) {
  
    String[] codes = new String[ BYTES ];                   // tabella dei codici di Huffman
    
    fillTable( root, "", codes );                           // compilazione della tabella
    
    return codes;
  }
  
  // Compilazione della tabella tramite visita ricorsiva dell'albero di Huffman
  // n:     nodo visitato
  // code:  codice binario del percorso dalla radice al nodo n
  // codes: variabile di stato per registrare i codici dei caratteri
  
  private static void fillTable( Node n, String code, String[] codes ) {
  
    if ( n.isLeaf() ) {
      codes[ n.getByte() ] = code;                          // nodo foglia: code = codice del carattere
    } else {
      fillTable( n.left(),  code+"0", codes );              // figlio sinistro: bit 0 nel codice
      fillTable( n.right(), code+"1", codes );              // figlio destro:   bit 1 nel codice
    }
  }
  
  
  // 4.
  // Codifica lineare dell'albero di Huffman tramite visita ricorsiva
  // n: nodo visitato
  // Struttura:
  // - una foglia è codificata dal byte che rappresenta
  //   (i byte/caratteri speciali '@' e '\' sono preceduti da '\')
  // - un albero con più di un nodo è codificato
  //   linearizzando i sottoalberi sinistro e destro,
  //   quindi giustapponendo il simbolo '@' e le stringhe ottenute
  
  public static String flattenTree( Node n ) {
    
    if ( n.isLeaf() ) {                                     // foglia: codifica del byte
      int b = n.getByte();
      if ( (b == '\\') || (b == '@') ) {
        return ( "\\" + (char) b );                         // byte/caratteri speciali: \, @
      } else {
        return ( "" + (char) b );                           // altri byte
      }
    } else {
      return ( "@"                                          // @
             + flattenTree( n.left() )                      // codifica lineare del sottoalbero sinistro
             + flattenTree( n.right() )                     // codifica lineare del sottoalbero destro
               );
    }
  }
  
  
  // 5.
  // Compressione
  // src: nome del documento originale
  // dst: nome del documento compresso
  
  public static void compress( String src, String dst ) {
    
    // Scansione I : frequenze dei carattteri
    
    int[] freq = byteHistogram( src );                      // istogramma delle frequenze
    
    Node root = huffmanTree( freq );                        // albero di Huffman
    
    String tree = flattenTree(root);                        // albero linearizzato
    
    int count = root.weight();                              // numero complessivo di byte
    String[] codes = huffmanCodesTable( root );             // tabella (array) dei codici di Huffman
    
    // Scansione II : codifica di Huffman
    
    InputByteFile in = new InputByteFile( src );            // documento originale
    
    OutputByteFile out = new OutputByteFile( dst );         // documento compresso
    
    writeHeader( count, tree, out );                        // intestazione del documento compresso
    
    for ( int j=0; j<count; j=j+1 ) {                       // scansione: codifica dei caratteri
    
      int b = in.readByte();
      out.writeCode( codes[b] );
    }
    in.close();
    out.close();
  }
  
  private static void writeHeader( int count, String htree, OutputByteFile out ) {
  
    int[] digit = new int[ 4 ];                             // lunghezza del documento < maxint
    
    for ( int i=0; i<4; i=i+1 ) {
    
      digit[i] = count % BYTES;
      count = count / BYTES;
    }
    for ( int i=3; i>=0; i=i-1 ) {
    
      out.writeByte( digit[i] );
    }
    out.writeByte( '\n' );
    
    for ( int i=0; i<htree.length(); i=i+1 ) {
    
      out.writeByte( htree.charAt(i) );
    }
    out.writeByte( '\n' );
  }
  
  
}  // class HuffmanCompression

